from typing import List, Tuple, Dict
import pytest


from tealer.teal.basic_blocks import BasicBlock
from tealer.teal.instructions import instructions
from tealer.teal.instructions import transaction_field
from tealer.teal import global_field
from tealer.utils.command_line.common import init_tealer_from_single_contract

from tests.utils import cmp_cfg, construct_cfg, order_basic_blocks


MULTIPLE_RETSUB = """
#pragma version 5
b main
is_even:
    int 2
    %
    bz return_1
    int 0
    global GroupSize
    int 2
    ==
    assert
    retsub
return_1:
    global GroupSize
    int 3
    <
    assert
    int 1
    retsub
main:
    global GroupSize
    int 1
    !=
    assert
    int 4
    callsub is_even
    return
"""

ins_list = [
    instructions.Pragma(5),
    instructions.B("main"),
    instructions.Label("is_even"),
    instructions.Int(2),
    instructions.Modulo(),
    instructions.BZ("return_1"),
    instructions.Int(0),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(2),
    instructions.Eq(),
    instructions.Assert(),
    instructions.Retsub(),
    instructions.Label("return_1"),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(3),
    instructions.Less(),
    instructions.Assert(),
    instructions.Int(1),
    instructions.Retsub(),
    instructions.Label("main"),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(1),
    instructions.Neq(),
    instructions.Assert(),
    instructions.Int(4),
    instructions.Callsub("is_even"),
    instructions.Return(),
]

ins_partitions = [(0, 2), (2, 6), (6, 12), (12, 19), (19, 26), (26, 27)]
bbs_links = [(0, 4), (4, 5), (1, 2), (1, 3)]

MULTIPLE_RETSUB_CFG_GROUP_SIZES = [[2], [2], [2], [2], [2], [2]]

MULTIPLE_RETSUB_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)


SUBROUTINE_BACK_JUMP = """
#pragma version 5
b main
getmod:
    %
    retsub
is_odd:
    global GroupSize
    int 4
    <
    assert
    global GroupSize
    int 2
    !=
    assert
    int 2
    b getmod
main:
    int 5
    callsub is_odd
    return
"""

ins_list = [
    instructions.Pragma(5),
    instructions.B("main"),
    instructions.Label("getmod"),
    instructions.Modulo(),
    instructions.Retsub(),
    instructions.Label("is_odd"),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(4),
    instructions.Less(),
    instructions.Assert(),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(2),
    instructions.Neq(),
    instructions.Assert(),
    instructions.Int(2),
    instructions.B("getmod"),
    instructions.Label("main"),
    instructions.Int(5),
    instructions.Callsub("is_odd"),
    instructions.Return(),
]

ins_partitions = [(0, 2), (2, 5), (5, 16), (16, 19), (19, 20)]
bbs_links = [(0, 3), (3, 4), (2, 1)]

SUBROUTINE_BACK_JUMP_CFG_GROUP_SIZES = [[1, 3], [1, 3], [1, 3], [1, 3], [1, 3], [1, 3]]
SUBROUTINE_BACK_JUMP_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)

BRANCHING = """
#pragma version 4
global GroupSize
int 2
>=
assert
global GroupSize
int 4
>
bz fin
global GroupSize
int 1
==
bnz check_second_arg
int 0
return
check_second_arg:
txn ApplicationArgs 1
btoi
int 100
>
bnz fin
int 0
return
fin:
int 1
return
"""

ins_list = [
    instructions.Pragma(4),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(2),
    instructions.GreaterE(),
    instructions.Assert(),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(4),
    instructions.Greater(),
    instructions.BZ("fin"),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(1),
    instructions.Eq(),
    instructions.BNZ("check_second_arg"),
    instructions.Int(0),
    instructions.Return(),
    instructions.Label("check_second_arg"),
    instructions.Txn(transaction_field.ApplicationArgs(1)),
    instructions.Btoi(),
    instructions.Int(100),
    instructions.Greater(),
    instructions.BNZ("fin"),
    instructions.Int(0),
    instructions.Return(),
    instructions.Label("fin"),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [(0, 9), (9, 13), (13, 15), (15, 21), (21, 23), (23, 26)]
bbs_links = [(0, 1), (0, 5), (1, 2), (1, 3), (3, 4), (3, 5)]

BRANCHING_CFG_GROUP_SIZES = [[2, 3, 4], [], [], [], [], [2, 3, 4]]
BRANCHING_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)

LOOPS = """
#pragma version 5
global GroupSize
int 4
!=
assert
int 0
loop:
    dup
    global GroupSize
    int 3
    >=
    bz end
    int 1
    +
    global GroupSize
    int 3
    <
    assert
    b loop
end:
    int 2
    global GroupSize
    ==
    assert
    int 1
    return
"""

ins_list = [
    instructions.Pragma(5),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(4),
    instructions.Neq(),
    instructions.Assert(),
    instructions.Int(0),
    instructions.Label("loop"),
    instructions.Dup(),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(3),
    instructions.GreaterE(),
    instructions.BZ("end"),
    instructions.Int(1),
    instructions.Add(),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(3),
    instructions.Less(),
    instructions.Assert(),
    instructions.B("loop"),
    instructions.Label("end"),
    instructions.Int(2),
    instructions.Global(global_field.GroupSize()),
    instructions.Eq(),
    instructions.Assert(),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [(0, 6), (6, 12), (12, 19), (19, 26)]
bbs_links = [(0, 1), (1, 2), (1, 3), (2, 1)]

LOOPS_CFG_GROUP_SIZES = [[2], [2], [], [2]]
LOOPS_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)

cfg_group_sizes = [
    (MULTIPLE_RETSUB_CFG, MULTIPLE_RETSUB_CFG_GROUP_SIZES),
    (SUBROUTINE_BACK_JUMP_CFG, SUBROUTINE_BACK_JUMP_CFG_GROUP_SIZES),
    (BRANCHING_CFG, BRANCHING_CFG_GROUP_SIZES),
    (LOOPS_CFG, LOOPS_CFG_GROUP_SIZES),
]

transaction_contexts = []
for cfg, sizes in cfg_group_sizes:
    bb = order_basic_blocks(cfg)
    transaction_context = {}
    for b, group_sizes in zip(bb, sizes):  # type: ignore
        transaction_context[b] = group_sizes
    transaction_contexts.append(transaction_context)

ALL_TESTS = [
    (MULTIPLE_RETSUB, MULTIPLE_RETSUB_CFG, transaction_contexts[0]),
    (SUBROUTINE_BACK_JUMP, SUBROUTINE_BACK_JUMP_CFG, transaction_contexts[1]),
    (BRANCHING, BRANCHING_CFG, transaction_contexts[2]),
    (LOOPS, LOOPS_CFG, transaction_contexts[3]),
]


@pytest.mark.parametrize("test", ALL_TESTS)  # type: ignore
def test_group_sizes(test: Tuple[str, List[BasicBlock], Dict]) -> None:
    code, cfg_tested, cfg_tested_txn_context = test

    tealer = init_tealer_from_single_contract(code.strip(), "test")
    function = tealer.contracts["test"].functions["test"]
    for bb_tested in cfg_tested:
        print(bb_tested)
    print("*" * 20)
    assert cmp_cfg(function.blocks, cfg_tested)

    bbs = order_basic_blocks(function.blocks)
    cfg_tested = order_basic_blocks(cfg_tested)
    for b1, b2 in zip(bbs, cfg_tested):
        print(function.transaction_context(b1).group_sizes, cfg_tested_txn_context[b2])
        assert function.transaction_context(b1).group_sizes == cfg_tested_txn_context[b2]


MULTIPLE_RETSUB_CFG_GROUP_INDICES = [[0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1]]
SUBROUTINE_BACK_JUMP_GROUP_INDICES = [
    [0, 1, 2],
    [0, 1, 2],
    [0, 1, 2],
    [0, 1, 2],
    [0, 1, 2],
    [0, 1, 2],
]
BRANCHING_GROUP_INDICES = [[0, 1, 2, 3], [], [], [], [], [0, 1, 2, 3]]
LOOPS_GROUP_INDICES = [[0, 1], [0, 1], [], [0, 1]]

GROUP_INDICES_TESTS = [
    (MULTIPLE_RETSUB, MULTIPLE_RETSUB_CFG_GROUP_INDICES),
    (SUBROUTINE_BACK_JUMP, SUBROUTINE_BACK_JUMP_GROUP_INDICES),
    (BRANCHING, BRANCHING_GROUP_INDICES),
    (LOOPS, LOOPS_GROUP_INDICES),
]


@pytest.mark.parametrize("test", GROUP_INDICES_TESTS)  # type: ignore
def test_group_indices(test: Tuple[str, List[List[int]]]) -> None:
    code, group_indices_list = test
    tealer = init_tealer_from_single_contract(code.strip(), "test")
    function = tealer.contracts["test"].functions["test"]

    bbs = order_basic_blocks(function.blocks)
    for bb_tested, group_indices in zip(bbs, group_indices_list):
        print(function.transaction_context(bb_tested).group_indices, group_indices)
        assert function.transaction_context(bb_tested).group_indices == group_indices



"""Printer to output information about transaction fields in the CFG."""

import os
from pathlib import Path
from typing import List, TYPE_CHECKING

from tealer.printers.abstract_printer import AbstractPrinter
from tealer.utils.output import (
    CFGDotConfig,
    full_cfg_to_dot,
    all_subroutines_to_dot,
    ROOT_OUTPUT_DIRECTORY,
)

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock


class PrinterTransactionContext(AbstractPrinter):  # pylint: disable=too-few-public-methods
    """Printer to output information about transaction fields in the CFG.

    Transaction field analysis finds possible values for transaction fields. This printer
    adds this information on top of each of the basic blocks in the output CFG.
    """

    NAME = "transaction-context"
    # TODO: Add TxnType, Sender, RekeyTo, CloseRemainderTo, AssetCloseTo field values
    HELP = "Output possible values of GroupIndices, GroupSize"
    WIKI_URL = "https://github.com/crytic/tealer/wiki/Printer-documentation#transaction-context"

    @staticmethod
    def _repr_num_list(values: List[int]) -> str:
        """Return short string representation of range of integers.

        intergers are space-separated and represented in ascending order.

        Continous sequence of more than 3 integers are represented using short-form(a..b).
        e.g
            5 6 7 8
           => 5..8

            1 2 3 5 6 7 8 9 11 13 14 15 16
           => 1 2 3 5..9 11 13..16

        Args:
            values: Sorted list of integers. smaller value is first.

        Returns:
            Returns short string representation of the :values:
        """
        values = sorted(values)
        sequences: List[List[int]] = [[]]
        for i in values:
            if not sequences[-1]:
                sequences[-1].append(i)
            elif sequences[-1][-1] == (i - 1):
                sequences[-1].append(i)
            else:
                sequences.append([i])
        str_seqs = []
        for seq in sequences:
            if len(seq) >= 4:
                str_seqs.append(f"{seq[0]}..{seq[-1]}")
            else:
                str_seqs.append(" ".join(str(i) for i in seq))
        return " ".join(str_seqs)

    def print(self) -> None:
        filename = Path("transaction-context.dot")

        # outputs multiple files: set the dir to {ROOT_DIRECTORY}/{CONTRACT_NAME}/{"print-"PRINTER_NAME}
        dest = ROOT_OUTPUT_DIRECTORY / Path(self.teal.contract_name) / Path(f"print-{self.NAME}")
        os.makedirs(dest, exist_ok=True)

        filename = dest / filename
        function = list(self.teal.functions.values())[0]

        def get_info(bb: "BasicBlock") -> List[str]:
            # NOTE: use the first function for now as `init_tealer_from_single_contract` uses entire contract as single function.
            group_indices_str = self._repr_num_list(function.transaction_context(bb).group_indices)
            group_sizes_str = self._repr_num_list(function.transaction_context(bb).group_sizes)
            return [f"GroupIndex: {group_indices_str}", f"GroupSize: {group_sizes_str}"]

        config = CFGDotConfig()
        config.bb_additional_comments = get_info
        # generate a Full CFG with all group-size, group-index comments
        full_cfg_to_dot(self.teal, config, filename)
        # Also generate shortened CFGs with group-size and group-index comments.
        all_subroutines_to_dot(
            self.teal, dest, config, "txn_ctx"
        )  # set prefix differentiate from other printers
        print(f"\nExported CFG with transaction context information to {filename}")



# pylint: skip-file
# mypy: ignore-errors
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import

from pyteal import *

from tealer.detectors.all_detectors import CanCloseAsset, CanCloseAccount

router = Router(
    name="Example",
    bare_calls=BareCallActions(),
)


@router.method(no_op=CallConfig.ALL)
def echo(input: abi.Uint64, *, output: abi.Uint64) -> Expr:
    """
    Method config validations Teal pattern:
        txn OnCompletion
        int NoOp
        ==
        txn ApplicationID
        int 0
        !=
        &&
        assert            // Assert(NoOp, CALL)

    Args:
        input: Input argument
        output: Return value of the echo method.

    Returns:
        Body of the echo method as PyTeal expr.
    """
    return output.set(input.get())


# pragma(compiler_version="0.22.0")
application_approval_program, _, _ = router.compile_program(
    version=7,
    assemble_constants=True,  # use intcblock, bytecblock
    # optimize=OptimizeOptions(scratch_slots=True),
)


@Subroutine(TealType.none)
def process_txn() -> Expr:
    return Pop(Bytes("ExecuteTransaction"))


def axfer_approval_program():
    @Subroutine(TealType.none)
    def validate_txn() -> Expr:
        return Seq(
            [
                Assert(Global.group_size() == Int(1)),
                Assert(Txn.fee() <= Global.min_txn_fee()),
                Assert(Txn.type_enum() == TxnType.AssetTransfer),
                Assert(Txn.asset_close_to() == Global.zero_address()),
            ]
        )

    return Seq(
        [
            validate_txn(),
            process_txn(),
            Return(Int(1)),
        ]
    )


axfer_approval_program_teal = compileTeal(axfer_approval_program(), mode=Mode.Signature, version=7)


def payment_approval_program():
    @Subroutine(TealType.none)
    def validate_txn() -> Expr:
        return Seq(
            [
                Assert(Global.group_size() == Int(1)),
                Assert(Txn.fee() <= Global.min_txn_fee()),
                Assert(Txn.type_enum() == TxnType.Payment),
                Assert(Txn.close_remainder_to() == Global.zero_address()),
            ]
        )

    return Seq(
        [
            validate_txn(),
            process_txn(),
            Return(Int(1)),
        ]
    )


payment_approval_program_teal = compileTeal(
    payment_approval_program(), mode=Mode.Signature, version=7
)


def axfer_payment_approval_program():
    return (
        If(Arg(Int(0)) == Bytes("Payment In Algo"))
        .Then(
            payment_approval_program(),
        )
        .Else(
            axfer_approval_program(),
        )
    )


axfer_payment_approval_program_teal = compileTeal(
    axfer_payment_approval_program(), mode=Mode.Signature, version=7
)


def different_group_sizes_approval_program():
    group_size_3 = Seq(
        [
            Assert(Global.group_size() == Int(3)),
            Assert(Gtxn[0].fee() <= Global.min_txn_fee()),
            Assert(Gtxn[0].type_enum() == TxnType.Payment),
            Assert(Gtxn[0].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[1].fee() <= Global.min_txn_fee()),
            Assert(Gtxn[1].type_enum() == TxnType.KeyRegistration),
            Assert(Gtxn[2].fee() <= Global.min_txn_fee()),
            Assert(Gtxn[2].type_enum() == TxnType.AssetConfig),
            Assert(Gtxn[2].close_remainder_to() == Global.zero_address()),
            Return(Int(1)),
        ]
    )

    group_size_2 = Seq(
        [
            Assert(Global.group_size() == Int(2)),
            Assert(Gtxn[0].fee() <= Global.min_txn_fee()),
            Assert(Gtxn[0].type_enum() == TxnType.AssetTransfer),
            Assert(Gtxn[0].asset_close_to() == Txn.sender()),
            Assert(Gtxn[1].fee() <= Global.min_txn_fee()),
            Assert(Gtxn[1].type_enum() == TxnType.AssetFreeze),
            Return(Int(1)),
        ]
    )

    return (
        If(Arg(Int(0)) == Bytes("First Operation"))
        .Then(
            group_size_2,
        )
        .Else(group_size_3)
    )


different_group_sizes_approval_program_teal = compileTeal(
    different_group_sizes_approval_program(), mode=Mode.Signature, version=7
)


txn_type_based_tests = [
    (application_approval_program, CanCloseAccount, []),
    (application_approval_program, CanCloseAsset, []),
    (axfer_approval_program_teal, CanCloseAccount, []),
    (axfer_approval_program_teal, CanCloseAsset, []),
    (payment_approval_program_teal, CanCloseAccount, []),
    (payment_approval_program_teal, CanCloseAsset, []),
    (axfer_payment_approval_program_teal, CanCloseAccount, []),
    (axfer_payment_approval_program_teal, CanCloseAsset, []),
    (different_group_sizes_approval_program_teal, CanCloseAccount, []),
    (different_group_sizes_approval_program_teal, CanCloseAsset, []),
]



"""Defines exceptions classes for representing exceptions in tealer.

* TealerException: Exception class for common exceptions in tealer.
"""


class TealerException(Exception):
    """Exception class to represent common exceptions in tealer"""



"""Parser for fields of asset_params_get.

Each ``asset_params_get`` field is represented as a class. Parsing
the field is creating the class instance representing the field given
it's string representation.

Fields of ``asset_params_get`` doesn't have immediate arguments
and have very simple structure. All fields map one-to-one with
their string representation. As a result, asset parameter fields are
parsed by first constructing a map(dict) from their string
representation to corresponding class representing the field and
using that to find the correct class for the given field.

Attributes:
    ASSET_PARAMS_FIELD_TXT_TO_OBJECT: Map(dict) from string representation
        of asset parameter field to the corresponding class.
"""

from tealer.teal.instructions.asset_params_field import (
    AssetParamsField,
    AssetTotal,
    AssetDecimals,
    AssetDefaultFrozen,
    AssetUnitName,
    AssetName,
    AssetURL,
    AssetMetadataHash,
    AssetManager,
    AssetReserve,
    AssetFreeze,
    AssetClawback,
    AssetCreator,
)

ASSET_PARAMS_FIELD_TXT_TO_OBJECT = {
    "AssetTotal": AssetTotal,
    "AssetDecimals": AssetDecimals,
    "AssetDefaultFrozen": AssetDefaultFrozen,
    "AssetUnitName": AssetUnitName,
    "AssetName": AssetName,
    "AssetURL": AssetURL,
    "AssetMetadataHash": AssetMetadataHash,
    "AssetManager": AssetManager,
    "AssetReserve": AssetReserve,
    "AssetFreeze": AssetFreeze,
    "AssetClawback": AssetClawback,
    "AssetCreator": AssetCreator,
}


def parse_asset_params_field(field: str) -> AssetParamsField:
    """Parse fields of ``asset_params_get``.

    Args:
        field: string representation of the field.

    Returns:
        object of class corresponding to the given asset_params_get field.
    """

    return ASSET_PARAMS_FIELD_TXT_TO_OBJECT[field]()



"""Detector for finding execution paths missing RekeyTo check."""

from typing import List, TYPE_CHECKING, Tuple

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.detectors.utils import (
    detect_missing_tx_field_validations_group,
    detect_missing_tx_field_validations_group_complete,
)
from tealer.utils.output import ExecutionPaths


if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.utils.output import ListOutput, GroupTransactionOutput
    from tealer.teal.context.block_transaction_context import BlockTransactionContext
    from tealer.teal.teal import Teal


class MissingRekeyTo(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find execution paths missing RekeyTo check.

    TEAL, from version 2 onwards supports rekeying of accounts.
    An account can be rekeyed to a different address. Once rekeyed,
    rekeyed address has entire authority over the account. Contract
    Accounts can also be rekeyed. If RekeyTo field of the transaction
    is set to malicious actor's address, then they can control the account
    funds, assets directly bypassing the contract's restrictions.

    This detector tries to find execution paths that approve the algorand
    transaction("return 1") and doesn't check the RekeyTo transaction field.
    Additional to checking rekeying of it's own contract, detector also finds
    execution paths that doesn't check RekeyTo field of other transactions
    in the atomic group.
    """

    NAME = "rekey-to"
    DESCRIPTION = "Rekeyable Logic Signatures"
    TYPE = DetectorType.STATELESS

    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = "https://github.com/crytic/tealer/wiki/Detector-Documentation#rekeyable-logicsig"
    WIKI_TITLE = "Rekeyable LogicSig"
    WIKI_DESCRIPTION = (
        "Logic signature does not validate `RekeyTo` field."
        " Attacker can submit a transaction with `RekeyTo` field set to their address and take control over the account."
        " More at [building-secure-contracts/not-so-smart-contracts/algorand/rekeying]"
        "(https://github.com/crytic/building-secure-contracts/tree/master/not-so-smart-contracts/algorand/rekeying)"
    )
    WIKI_EXPLOIT_SCENARIO = """
```py
def withdraw(...) -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.type_enum() == TxnType.Payment,
                    Txn.first_valid() % period == Int(0),
                    Txn.last_valid() == Txn.first_valid() + duration,
                    Txn.receiver() == receiver,
                    Txn.amount() == amount,
                    Txn.first_valid() < timeout,
                )
            ),
            Approve(),
        ]
    )
```

Alice signs the logic-sig to allow recurring payments to Bob.\
 Eve uses the logic-sig and submits a valid transaction with `RekeyTo` field set to her address.\
 Eve takes over Alice's account.
"""

    WIKI_RECOMMENDATION = """
Validate `RekeyTo` field in the LogicSig.
"""

    def detect(self) -> "ListOutput":
        """Detect execution paths with missing CloseRemainderTo check.

        Returns:
            ExecutionPaths instance containing the list of vulnerable execution
            paths along with name, check, impact, confidence and other detector
            information.
        """

        def checks_field(block_ctx: "BlockTransactionContext") -> bool:
            # return False if RekeyTo field can have any address.
            # return True if RekeyTo should have some address or zero address
            return not block_ctx.rekeyto.any_addr

        # there should be a better to decide which function to call ??
        if self.tealer.output_group:
            # mypy complains if the value is returned directly. Uesd the second suggestion mentioned here:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#variance
            return list(
                detect_missing_tx_field_validations_group_complete(self.tealer, self, checks_field)
            )

        output: List[
            Tuple["Teal", List[List["BasicBlock"]]]
        ] = detect_missing_tx_field_validations_group(self.tealer, checks_field)
        detector_output: "ListOutput" = []
        for contract, vulnerable_paths in output:
            detector_output.append(ExecutionPaths(contract, self, vulnerable_paths))

        return detector_output



# pylint: skip-file
# mypy: ignore-errors
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import
from tealer.detectors.all_detectors import IsUpdatable

# Basic recursion
@Subroutine(TealType.uint64)  # type: ignore[misc]
def factorial(n: Expr) -> Expr:
    # Note: Recursion is not supported if the subroutine takes ScratchVar type argument
    return If(
        n == Int(0),  # condition
        Return(Int(1)),  # then expression
        Return(n * factorial(n - Int(1))),  # else expression
    )


# recursion with multiple subroutines
@Subroutine(TealType.none)
def recursive_1(n: Expr) -> Expr:
    return Seq(
        [
            Pop(Bytes("Recursive 1 Subroutine")),
            If(
                n < Int(5),
                recursive_2(n),
            ),
        ]
    )


@Subroutine(TealType.none)
def recursive_2(n: Expr) -> Expr:
    return Seq([recursive_3(n)])


@Subroutine(TealType.none)
def recursive_3(n: Expr) -> Expr:
    return Seq(
        [
            Pop(Bytes("Recursive 2 Subroutine")),
            recursive_1(n + Int(1)),
            Return(),
        ]
    )


# loop in a subroutine
@Subroutine(TealType.none)
def loop_in_sub() -> Expr:
    i = ScratchVar()
    return Seq(
        [
            Pop(Bytes("Loop in Subroutine")),
            For(i.store(Int(0)), i.load() < Int(10), i.store(i.load() + Int(1)),).Do(
                Log(Itob(i.load())),
            ),
            Return(),
        ]
    )


# loop in a subroutine calling a recursive subroutine
@Subroutine(TealType.none)
def loop_recursive_1(n: Expr) -> Expr:
    i = ScratchVar()
    return Seq(
        [
            Pop(Bytes("Loop Recursive Subroutine")),
            For(i.store(Int(0)), i.load() < Int(2), i.store(i.load() + Int(1)),).Do(
                If(
                    n < Int(3),
                    loop_recursive_2(n),
                    Return(),
                )
            ),
        ]
    )


@Subroutine(TealType.none)
def loop_recursive_2(n: Expr) -> Expr:
    return Seq([loop_recursive_1(n + Int(1)), Return()])


def compile_program(approval_program: Expr) -> str:
    return compileTeal(
        approval_program,
        mode=Mode.Application,
        version=7,
    )


basic_recursion = Seq([Pop(factorial(Int(4))), Approve()])
multiple_recursion = Seq([recursive_1(Int(1)), Approve()])
loop_in_subroutine = Seq([loop_in_sub(), Approve()])
loop_and_recursion = Seq([loop_recursive_1(Int(0)), Approve()])

multiple_calls_to_same_sub = Seq([loop_in_sub(), loop_in_sub(), Approve()])

all_patterns = Seq(
    [
        Pop(factorial(Int(4))),
        recursive_1(Int(1)),
        loop_in_sub(),
        loop_recursive_1(Int(0)),
        loop_in_sub(),
        Approve(),
    ]
)

basic_recursion_teal = compile_program(basic_recursion)

multiple_recursion_teal = compile_program(multiple_recursion)
loop_in_subroutine_teal = compile_program(loop_in_subroutine)
loop_and_recursion_teal = compile_program(loop_and_recursion)

multiple_calls_to_same_sub_teal = compile_program(multiple_calls_to_same_sub)

all_patterns_teal = compile_program(all_patterns)


# Test that detectors are enumerating paths correctly by asserting the detector results.
# None of the programs contain a check for any of the detectors. So, detectors should report
# all valid paths.

# TODO: Having a edge from callsub to the called subroutine was a really bad idea. Change this as soon as
# possible.

basic_recursion_paths = [
    [0, 2, 5, 1],
    # We completely ignore the path using the recursion/loop.
    # May be we should consider one iteration and generate path from that
    # [0,
    # 2, 3, 4,  # Block 4 calls subroutine in Block 2
    # 2, 5
    # 1]
]

multiple_recursion_paths = [
    [0, 2, 5, 1],
    # [0, 2, 3, 6, 8, 2, 5, 1]
]

loop_in_subroutine_paths = [
    [0, 2, 3, 5, 1],
    # [0, 2, 3, 4, 3, 5, 1]
]

loop_and_recursion_paths = [
    [0, 2, 3, 4, 5, 1],
    [0, 2, 3, 9, 1],
    ### [0, 2, 3, 4, 7, 10, 2, .., 11, 8, 6, 3, ...]
    # [0, 2, 3, 4, 7, 10, 2, 3, 9, 1]
    # [0, 2, 3, 4, 7, 10, 2, 3, 4, 5, 1]
    # [0, 2, 3, 4, 7, 10, 2, 3, 9, 11, 8, 6, 3, 9, 1]
    # [0, 2, 3, 4, 7, 10, 2, 3, 4, 5, 11, 8, 6, 3, 9, 1]
    # [0, 2, 3, 4, 7, 10, 2, 3, 9, 11, 8, 6, 3, 4, 5, 1]
    # [0, 2, 3, 4, 7, 10, 2, 3, 4, 5, 11, 8, 6, 3, 4, 5, 1]
]

multiple_calls_to_same_sub_paths = [
    [0, 3, 4, 6, 1, 3, 4, 6, 2],
    # ...
]

all_patterns_paths = [
    [
        0,
        6,
        9,  # factorial
        1,
        10,
        13,  # recursive1_1
        2,
        18,
        19,
        21,  # loop in sub
        3,
        22,
        23,
        24,
        25,  # loop recursive
        4,
        18,
        19,
        21,  # loop in sub
        5,
    ],
    [
        0,
        6,
        9,  # factorial
        1,
        10,
        13,  # recursive1_1
        2,
        18,
        19,
        21,  # loop in sub
        3,
        22,
        23,
        29,  # loop recursive
        4,
        18,
        19,
        21,  # loop in sub
        5,
    ],
]


subroutine_recursion_patterns_tests = [
    (basic_recursion_teal, IsUpdatable, basic_recursion_paths),
    (multiple_recursion_teal, IsUpdatable, multiple_recursion_paths),
    (loop_in_subroutine_teal, IsUpdatable, loop_in_subroutine_paths),
    (loop_and_recursion_teal, IsUpdatable, loop_and_recursion_paths),
    (multiple_calls_to_same_sub_teal, IsUpdatable, multiple_calls_to_same_sub_paths),
    (all_patterns_teal, IsUpdatable, all_patterns_paths),
]

if __name__ == "__main__":
    print(basic_recursion_teal)
    print(multiple_recursion_teal)
    print(loop_in_subroutine_teal)
    print(loop_and_recursion_teal)
    print(multiple_calls_to_same_sub_teal)
    print(all_patterns_teal)



"""Detector for finding unoptimized self access"""

from typing import TYPE_CHECKING, List

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.teal.instructions.instructions import Gtxns, Instruction, Txn, Gtxnsa, Gtxnsas
from tealer.teal.instructions.transaction_field import GroupIndex
from tealer.utils.output import InstructionsOutput

if TYPE_CHECKING:
    from tealer.utils.output import ListOutput


class SelfAccess(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find unoptimized self access"""

    NAME = "self-access"
    DESCRIPTION = "Unoptimized self access"
    TYPE = DetectorType.STATELESS

    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = (
        "https://github.com/crytic/tealer/wiki/Detector-Documentation#Unoptimized-self-access"
    )
    WIKI_TITLE = "Unoptimized self access"
    WIKI_DESCRIPTION = "Gtxn[Txn.group_index()].field can be replaced by Txn.field"
    WIKI_EXPLOIT_SCENARIO = """"""

    WIKI_RECOMMENDATION = """Use Txn.field."""

    def detect(self) -> "ListOutput":
        """Detector to find unoptimized usage of self access

        Returns:
              The two instructions to optimized
        """

        detector_output: "ListOutput" = []

        for contract in self.tealer.contracts.values():

            all_findings: List[List[Instruction]] = []

            for function in contract.functions.values():
                for block in function.blocks:
                    if len(block.instructions) < 2:
                        continue

                    # Sliding windows
                    for i in range(0, len(block.instructions) - 1):
                        first = block.instructions[i]
                        second = block.instructions[i + 1]

                        # Note: first / second are always assigned, as we check for len(block.instructions) first
                        # pylint: disable=undefined-loop-variable
                        if (
                            isinstance(first, Txn)
                            and isinstance(first.field, GroupIndex)
                            and isinstance(second, (Gtxns, Gtxnsa, Gtxnsas))
                        ):
                            # pylint: disable=undefined-loop-variable
                            all_findings.append([first, second])

            if all_findings:
                detector_output.append(InstructionsOutput(contract, self, all_findings))

        return detector_output



"""
For a given field multiple types of possible values can be calculated.
The possible values of the field of the transaction
    1. executing the contract/function irrespective of the transaction's group index.
    2. executing the contract when the transaction group index is a particular value.
    3. at a given absolute index irrespective of which contract is executed.
    4. at a given relative index irrespective of which contract is executed.

see transaction_context/__init__.py for more information.
"""

from typing import TYPE_CHECKING, Tuple, Optional, Type

from tealer.teal.instructions.parse_transaction_field import TX_FIELD_TXT_TO_OBJECT
from tealer.analyses.dataflow.transaction_context.utils.group_helpers import (
    IndexType,
    get_index_and_field,
)

if TYPE_CHECKING:
    from tealer.analyses.utils.stack_ast_builder import KnownStackValue
    from tealer.teal.instructions.transaction_field import TransactionField


def get_gtxn_at_index_key(idx: int, base_key: str) -> str:
    """Analysis to represent type (2) information for base_key field

    Args:
        idx: index of the transaction in the group (0 to MAX_GROUP_SIZE - 1)
        base_key: The key used to represent type (1) information of the field.

    Returns:
        Returns the key used to represent type (2) information in the analysis
    """
    return f"GTXN_AT_INDEX_{idx:02d}_{base_key}"


def is_gtxn_at_index_key(analysis_key: str) -> bool:
    """Return True if the analysis key represents type (2) information

    Args:
        analysis_key: The key.

    Returns:
        Returns True if analysis_key starts with "GTXN_AT_INDEX" or else False.
    """
    return analysis_key.startswith("GTXN_AT_INDEX")


def get_ind_base_for_gtxn_type_keys(analysis_key: str) -> Tuple[int, str]:
    """Calculate the index and base key for the txn_at_index, absolute, relative type keys.

    Args:
        analysis_key: A key

    Returns:
        Returns the transaction index and the base key for the analysis_key.
    """
    # will fail if the key is not a TXN_AT_INDEX key.
    ind, base_key = analysis_key.split("_")[-2:]
    return int(ind), base_key


def get_absolute_index_key(idx: int, base_key: str) -> str:
    """Analysis to represent type (3) information for base_key field

    Args:
        idx: index of the transaction in the group (0 to MAX_GROUP_SIZE - 1)
        base_key: The key used to represent type (1) information of the field.

    Returns:
        Returns the key used to represent type (3) information in the analysis
    """
    return f"GTXN_ABS_{idx:02d}_{base_key}"


def is_absolute_index_key(analysis_key: str) -> bool:
    """Return True if the analysis key represents type (3) information

    Args:
        analysis_key: The key.

    Returns:
        Returns True if analysis_key is a absolute type of key.
    """
    return analysis_key.startswith("GTXN_ABS_")


def get_relative_index_key(offset: int, base_key: str) -> str:
    """Analysis to represent type (2) information for base_key field

    Args:
        offset: index of the transaction in the group  (-(MAX_GROUP_SIZE - 1) to (MAX_GROUP - 1))
        base_key: The key used to represent type (1) information of the field.

    Returns:
        Returns the key used to represent type (4) information in the analysis
    """
    return f"GTXN_RELATIVE_{offset:02d}_{base_key}"


def is_relative_index_key(analysis_key: str) -> bool:
    """Return True if the analysis key represents type (4) information

    Args:
        analysis_key: The key.

    Returns:
        Returns True if analysis_key is a relative type of key.
    """
    return analysis_key.startswith("GTXN_RELATIVE_")


# pylint: disable=too-many-branches
def is_value_matches_key(
    analysis_key: str,
    stack_value: "KnownStackValue",
    key_field: Optional[Type["TransactionField"]] = None,
) -> bool:
    """return True if the stack_value is value of the field tracked by the analysis_key else False.

    Type of keys:
        Self: The value of the current transaction irrespective of its index in the group
        GTXN_AT_INDEX: The value of the transaction field executing this contract when this transaction is at the index.
        Absolute: The field value of the transaction at an absolute index irrespective of the contract being executed at that index.
        Relative: The field value of the transaction at an offset from the current transaction irrespective of the contract being executed at that offset.

    Args:
        analysis_key: The key tracking a field
        stack_value: The value currently being checked
        key_field: The transaction field if it cannot be calculated from the analysis_key

    Returns:
        Returns True if the stack_value is the value of the field tracked by the analysis_key else False.
    """
    value_is_field, value_index, value_field = get_index_and_field(stack_value)
    if not value_is_field:
        return False
    assert value_index is not None and value_field is not None
    if value_index.index_type == IndexType.Unknown:
        return False

    if key_field is None:
        if (
            is_gtxn_at_index_key(analysis_key)
            or is_absolute_index_key(analysis_key)
            or is_relative_index_key(analysis_key)
        ):

            _, field_str = get_ind_base_for_gtxn_type_keys(analysis_key)
        else:
            field_str = analysis_key
        field = TX_FIELD_TXT_TO_OBJECT[field_str]
    else:
        field = key_field

    if not isinstance(value_field, field):
        return False

    if is_gtxn_at_index_key(analysis_key) or is_absolute_index_key(analysis_key):
        # difference between Gtxn_at_key and absolute_index comes from merging of Type(1) information before propagating.
        # It is handled in DataflowTransactionContext._update_gtxn_constraints
        if value_index.index_type != IndexType.Absolute:
            return False

        idx, _ = get_ind_base_for_gtxn_type_keys(analysis_key)
        if value_index.value == idx:
            return True
        return False

    if is_relative_index_key(analysis_key):
        if value_index.index_type != IndexType.Relative:
            return False
        offset, _ = get_ind_base_for_gtxn_type_keys(analysis_key)
        if value_index.value == offset:
            return True
        return False

    # Txn/Self key
    if value_index.index_type == IndexType.Self:
        return True
    return False



"""Detector for finding execution paths missing Fee check."""

from typing import List, TYPE_CHECKING, Tuple

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.detectors.utils import (
    detect_missing_tx_field_validations_group,
    detect_missing_tx_field_validations_group_complete,
)
from tealer.utils.algorand_constants import MAX_TRANSACTION_COST
from tealer.utils.output import ExecutionPaths

if TYPE_CHECKING:
    from tealer.utils.output import ListOutput
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.context.block_transaction_context import BlockTransactionContext
    from tealer.teal.teal import Teal


class MissingFeeCheck(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find execution paths missing Fee check.

    The fee for stateless contract transactions will be deducted
    from the contract account or the LogicSig signer account. An
    attacker could set the fee to high value and drain the account
    funds in form of fees.

    This detector tries to find execution paths that approve the algorand
    transaction("return 1") and doesn't check the Fee field.
    """

    NAME = "missing-fee-check"
    DESCRIPTION = "Missing Fee Field Validation"
    TYPE = DetectorType.STATELESS

    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = (
        "https://github.com/crytic/tealer/wiki/Detector-Documentation#missing-fee-field-validation"
    )
    WIKI_TITLE = "Missing Fee Field Validation"
    WIKI_DESCRIPTION = (
        "LogicSig does not validate `Fee` field."
        " Attacker can submit a transaction with `Fee` field set to large value and drain the account balance."
        " More at [building-secure-contracts/not-so-smart-contracts/algorand/unchecked_transaction_fee]"
        "(https://github.com/crytic/building-secure-contracts/tree/master/not-so-smart-contracts/algorand/unchecked_transaction_fee)"
    )
    WIKI_EXPLOIT_SCENARIO = """
```py
def withdraw(...) -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.type_enum() == TxnType.Payment,
                    Txn.first_valid() % period == Int(0),
                    Txn.last_valid() == Txn.first_valid() + duration,
                    Txn.receiver() == receiver,
                    Txn.amount() == amount,
                    Txn.first_valid() < timeout,
                )
            ),
            Approve(),
        ]
    )
```

Alice signs the logic-sig to allow recurring payments to Bob.\
 Eve uses the logic-sig and submits a valid transaction with `Fee` set to 1 million ALGOs.\
 Alice loses 1 million ALGOs.
"""

    WIKI_RECOMMENDATION = """
Validate `Fee` field in the LogicSig.
"""

    def detect(self) -> "ListOutput":
        """Detect execution paths with missing Fee check.

        Returns:
            ExecutionPaths instance containing the list of vulnerable execution
            paths along with name, check, impact, confidence and other detector
            information.
        """

        def checks_field(block_ctx: "BlockTransactionContext") -> bool:
            # returns True if fee is bounded by some unknown value
            # or is bounded by some known value less than maximum transaction cost.
            return block_ctx.max_fee_unknown or block_ctx.max_fee <= MAX_TRANSACTION_COST

        # there should be a better to decide which function to call ??
        if self.tealer.output_group:
            # mypy complains if the value is returned directly. Uesd the second suggestion mentioned here:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#variance
            return list(
                detect_missing_tx_field_validations_group_complete(self.tealer, self, checks_field)
            )

        output: List[
            Tuple["Teal", List[List["BasicBlock"]]]
        ] = detect_missing_tx_field_validations_group(self.tealer, checks_field)
        detector_output: "ListOutput" = []
        for contract, vulnerable_paths in output:
            detector_output.append(ExecutionPaths(contract, self, vulnerable_paths))

        return detector_output
